-------------------------------------------------------------------------------
--! @file       CipherCore_Datapath.vhd
--! @author     Ekawat (ice) Homsirikamol
--! @brief      Datapath unit for Deoxys
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.AEAD_pkg.all;
use work.AES_pkg.all;

entity CipherCore_Datapath is
    generic (
        G_MAX_LEN       : integer := SINGLE_PASS_MAX    
    );
    port    (
        clk             : in  std_logic;
        rst             : in  std_logic;

        --! Input
        bdi             : in  std_logic_vector(128                  -1 downto 0);
        key             : in  std_logic_vector(32                   -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(128/8                -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(128/8                -1 downto 0);

        --! Control
        addr            : in  std_logic_vector(4                    -1 downto 0);
        stage_status    : in  std_logic_vector(4                    -1 downto 0);
        ld_block        : in  std_logic;
        ld_tweakey      : in  std_logic;
        ld_chksum       : in  std_logic;
        ld_auth         : in  std_logic;
        ld_state        : in  std_logic;
        en_auth         : in  std_logic;
        en_block        : in  std_logic;
        en_tweakey      : in  std_logic;
        en_chksum       : in  std_logic;
        en_state        : in  std_logic;
        en_key          : in  std_logic;
        en_npub         : in  std_logic;
        sel_input       : in  std_logic_vector(2                    -1 downto 0);
        sel_chksum      : in  std_logic;
        sel_rdkey       : in  std_logic;
        sel_output      : in  std_logic;
        sel_decrypt     : in  std_logic;
        sel_npub        : in  std_logic;
        sel_tag         : in  std_logic;

        --! Output
        bdo             : out std_logic_vector(128                  -1 downto 0);
        tag             : out std_logic_vector(128                  -1 downto 0);
        msg_auth_valid  : out std_logic
    );
end entity CipherCore_Datapath;

architecture structure of CipherCore_Datapath is
    type     rcon_type  is array (0 to 15) of std_logic_vector(  7 downto 0);
    type     ram_type   is array (0 to 15) of std_logic_vector(127 downto 0);

    constant BLOCK_NO_WIDTH     : integer := G_MAX_LEN-7;
    constant rcon_rom           : rcon_type := (x"5E",x"BC",x"63",x"C6",
                                                x"97",x"35",x"6A",x"D4",
                                                x"B3",x"7D",x"FA",x"EF",
                                                x"C5",x"91",x"39",x"72");

    signal rdkey_ram            : ram_type;
    signal rcon                 : std_logic_vector(8                -1 downto 0);
    signal tweakey              : std_logic_vector(128*2            -1 downto 0);
    signal tweakey_r            : std_logic_vector(128*2            -1 downto 0);
    signal new_tweakey          : std_logic_vector(128*2            -1 downto 0);

    signal block_no             : std_logic_vector(BLOCK_NO_WIDTH   -1 downto 0);

    signal rdkey_first          : std_logic_vector(128              -1 downto 0);
    signal rdkey_first_s        : std_logic_vector(128              -1 downto 0);
    signal rdkey                : std_logic_vector(128              -1 downto 0);
    signal new_rdkey            : std_logic_vector(128              -1 downto 0);
    signal rdkey_sel            : std_logic_vector(128              -1 downto 0);

    signal checksum             : std_logic_vector(128              -1 downto 0);
    signal auth                 : std_logic_vector(128              -1 downto 0);
    signal from_round           : std_logic_vector(128              -1 downto 0);
    signal init_rdkey           : std_logic_vector(128              -1 downto 0);
    signal input_data           : std_logic_vector(128              -1 downto 0);
    signal state                : std_logic_vector(128              -1 downto 0);
    signal bdi_r                : std_logic_vector(128              -1 downto 0);
    signal key_r                : std_logic_vector(128              -1 downto 0);
    signal npub_r               : std_logic_vector(64               -1 downto 0);
    signal npub_sel             : std_logic_vector(64               -1 downto 0);
    signal output               : std_logic_vector(128              -1 downto 0);
    signal tagout               : std_logic_vector(128              -1 downto 0);

    signal bdi_vbytes_r         : std_logic_vector(128/8            -1 downto 0);
    signal bdi_ploc_r           : std_logic_vector(128/8            -1 downto 0);
    signal valid_bits           : std_logic_vector(128              -1 downto 0);
    signal pad_loc              : std_logic_vector(128              -1 downto 0);

    signal state_map            : t_AES_state;
    signal rdkey_map            : t_AES_state;
    signal rdkey_sel_map        : t_AES_state;
    signal from_round_enc_map   : t_AES_state;
    signal from_round_dec_map   : t_AES_state;
    signal from_round_map       : t_AES_state;

    constant ZEROS              : std_logic_vector(128                      -1 downto 0) := (others => '0');
begin
    prc:
    process(clk)
    begin
        if rising_edge(clk) then
            if (ld_block = '1') then
                block_no <= std_logic_vector(to_unsigned(0,BLOCK_NO_WIDTH));
            elsif (en_block = '1') then
                block_no <= std_logic_vector(
                    unsigned(block_no(BLOCK_NO_WIDTH-1 downto 0))
                    + to_unsigned(1,BLOCK_NO_WIDTH));
            end if;

            if (ld_tweakey = '1') then
                tweakey_r    <= tweakey;
                rdkey_first  <= rdkey_first_s;
            elsif (en_tweakey = '1') then
                tweakey_r    <= new_tweakey;
                rdkey_ram(to_integer(unsigned(addr))) <= new_rdkey;
            end if;

            if (ld_chksum = '1') then
                checksum    <= (others => '0');
            elsif (en_chksum = '1') then
                if sel_chksum = '1' then
                    checksum <= checksum xor ((output and valid_bits) xor pad_loc);
                else
                    checksum <= checksum xor bdi;
                end if;
            end if;

            if (ld_auth = '1') then
                auth        <= (others => '0');
            elsif (en_auth = '1') then
                auth        <= auth xor state;
            end if;

            if (ld_state = '1') then
                state           <= input_data xor init_rdkey;
                bdi_r           <= bdi;
                bdi_vbytes_r    <= bdi_valid_bytes;
                bdi_ploc_r      <= bdi_pad_loc;
            elsif (en_state = '1') then
                state           <= from_round;
            end if;

            if (en_key = '1') then
                key_r       <= key_r(95 downto 0) & key;
            end if;

            if (en_npub = '1') then
                npub_r      <= bdi(127 downto 64);
            end if;
        end if;
    end process;

    gInfo:
    for i in 15 downto 0 generate
        valid_bits(7+(i*8)    downto i*8) <= (others => bdi_vbytes_r(i));
        pad_loc   (7+(i*8)              ) <= bdi_ploc_r(i);
        pad_loc   (6+(i*8)    downto i*8) <= (others => '0');
    end generate;

    --! Tweakey
    npub_sel      <= npub_r when sel_npub = '1' else (others => '0');
    tweakey       <= key_r & stage_status & npub_sel
        & ZEROS(128-stage_status'length-npub_sel'length
                -BLOCK_NO_WIDTH-1 downto 0)
        & block_no;
    u_tweakey:
    entity work.TweakeyUpdate(structure)
    port map (ii => tweakey_r, oo => new_tweakey);

    --! State init
    with sel_input(1 downto 0) select
    input_data <= bdi               when "00",
                  (others => '0')   when "01",
                  checksum          when others;
    init_rdkey <= new_rdkey when sel_decrypt = '1' else rdkey_first;
    rdkey_sel  <= rdkey_first when sel_rdkey = '1' else rdkey;

    --! Round key
    rcon          <= rcon_rom(to_integer(unsigned(addr)));
    rdkey_first_s <= (    tweakey(128+128-1 downto 128+96) xor     tweakey(128-1 downto  96) xor x"01020408") &
                     (    tweakey(128+ 96-1 downto 128+64) xor     tweakey( 96-1 downto  64) xor x"2F2F2F2F") &
                     (    tweakey(128+ 64-1 downto 128+32) xor     tweakey( 64-1 downto  32)                ) &
                     (    tweakey(128+ 32-1 downto 128+ 0) xor     tweakey( 32-1 downto   0)                );
    new_rdkey     <= (new_tweakey(128+128-1 downto 128+96) xor new_tweakey(128-1 downto  96) xor x"01020408") &
                     (new_tweakey(128+ 96-1 downto 128+64) xor new_tweakey( 96-1 downto  64) xor (rcon & rcon & rcon & rcon)) &
                     (new_tweakey(128+ 64-1 downto 128+32) xor new_tweakey( 64-1 downto  32)                ) &
                     (new_tweakey(128+ 32-1 downto 128+ 0) xor new_tweakey( 32-1 downto   0)                );
    rdkey         <= rdkey_ram(to_integer(unsigned(addr)));

    --! Round
    u_map1: entity work.AES_map(structure)
    port map ( ii => state, oo => state_map);
    u_map2: entity work.AES_map(structure)
    port map ( ii => rdkey, oo => rdkey_map);
    u_map3: entity work.AES_map(structure)
    port map ( ii => rdkey_sel, oo => rdkey_sel_map);

    u_enc: entity work.AES_Round_Basic
    port map (
        din     => state_map,
        rkey    => rdkey_map,
        dout    => from_round_enc_map);
    u_dec: entity work.AES_InvRound_Basic
    port map (
        din     => state_map,
        rkey    => rdkey_sel_map,
        dout    => from_round_dec_map);

    from_round_map <= from_round_dec_map when sel_decrypt = '1'
                                         else from_round_enc_map;

    u_imap: entity work.AES_invmap(structure)
    port map ( ii => from_round_map, oo => from_round);

    --! Output
    output         <= state when sel_output = '1' else state xor bdi_r;
    tagout         <= state xor auth;
    bdo            <= tagout when sel_tag = '1' else output;

    --! Tag
    msg_auth_valid <= '1' when tagout = bdi else '0';
end architecture structure;